//////////////////////////////////////////////////////////////////////////////
//
//  NAME: Joon S.  Lee (jl)
//  FILE: PCEqn.C
//  PROJ: COKO
//  DATE: 10/9/96
//
// SYNOPSIS
//  Class definitions for Eqns.
//
//////////////////////////////////////////////////////////////////////////////
 
#ifndef _PCEQN_C_
#define _PCEQN_C_

#include "PCEqn.H"

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: CEqn
// Parameters: (char*, CPattern*)
//
// Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

CEqn::CEqn(char* lhs, CPattern* rhs):
  _rhs(NULL),
  _lhs(NULL),
  _evalFlag(0)
{
  if (lhs)
  	{
  	  _lhs = new char[strlen(lhs)+1];
  	  strcpy(_lhs, lhs);
  	}
  _rhs = rhs;
  _evalFlag = 0;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: ~CEqn 
// Parameters: (void)
//
// Destructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

CEqn::~CEqn(void)
{
  if (_lhs) { delete [] _lhs; _lhs = NULL; }
  if (_rhs) { delete _rhs; _rhs = NULL; }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: SetEF
// Parameters: (void)
//
// Set _evalFlag ( _evalFlag = 1 )
//         								    
//////////////////////////////////////////////////////////////////////////////

void CEqn::SetEF(void)
{
  _evalFlag = 1;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: UnsetEF
// Parameters: (void)
//
// Unset _evalFlag ( _evalFlag = 0 )
//         								    
//////////////////////////////////////////////////////////////////////////////

void CEqn::UnsetEF(void)
{
  _evalFlag = 0;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: Exec
// Parameters: (CState*)
//
// What to do when executed
//         								    
//////////////////////////////////////////////////////////////////////////////

CState* CEqn::Exec(CState * s)
{

  if (s->ISuccess())
	{
	  if (!_lhs)
		{
		  // Special case: Eqn is the form of "(current =) CPattern"
		  if (!s->CEmptyInvDep())
			{
			  cout << "Error @ CEqn : Exec" << endl;
			  cout << "current is already defined" << endl;
			  cout << "varable aliasing not yet allowed error" << endl;
			  exit(1);
			  s->SetSuccess(F);
			}
		  else
			{
			  if (_debug_)
				cout << "@CEqn::Exec - match for current start" << endl;
			  s->SetCurrentPattern(_rhs);
			  s->CurrentPattern()->SetStackPointer(s->IEnv()->getData());
			  s = _rhs->matchSTM(s);
			  if (s->ISuccess())
				{
				  if (_debug_)
					cout << "@CEqn::Exec - success match for current" << endl;
			  	  SetEF();
				  s = _rhs->DepListCheck(_lhs,s);	   
				}
			}
		}
	  else 
		{
		  if (!s->IEmptyInvDep(_lhs))
			{
			  cout << "Error @ CEqn : Exec " << endl;
			  cout << _lhs << " is already defined" << endl; 
			  cout << "varable aliasing not yet allowed error" << endl;
			  exit(1);
			  s->SetSuccess(F);
			}
		  else
			{
			  if (_debug_)
				cout << "@CEqn::Exec - match for " << _lhs << " start" << endl;

			  // Eqn is the form of "variable name = CPattern"
			  NonType* old_nontype = (s->Store())->Copy_kTree();
			  NonType* tmp_nontype = (s->IFindDeep(_lhs))->Child();
			  if (_debug_)
				{
				  cout << _lhs << "matched with ";
				  char* res;
				  cout << (res = tmp_nontype->representation()) << endl;
				  if(res) delete res;
				}

			  s->SetStore(tmp_nontype);
			  s->SetPatternDeep(_lhs, _rhs);
			  s = _rhs->matchSTM(s);
			  if (s->ISuccess())
				{
				  if (_debug_)
					cout << "@CEqn::Exec - success match for " << _lhs << endl;
				  SetEF();
				  s = _rhs->DepListCheck(_lhs,s);	   
				}
			  s->SetStore(old_nontype);

			  old_nontype = NULL; 
			  tmp_nontype = NULL;
			}
		}
	}
  return s;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: undoDEP
// Parameters: (CState*)
//
// undo any changes on dependency list
//         								    
//////////////////////////////////////////////////////////////////////////////

CState* CEqn::undoDEP(CState * s)
{
  if (_evalFlag)
	_rhs->undoDEP(_lhs,s);
  UnsetEF();
  return s;
}

#endif


