//////////////////////////////////////////////////////////////////
//
//  NAME: Joon S. Lee (jl)
//  
//  FILE: PCRule.C
//  PROJ: COKO
//  DATE: 6/96
// 
//  SYNOPSIS
//  This class is for Rule Declaration/Invoke statements.
//
/////////////////////////////////////////////////////////////////
#include "PCRule.H"

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: CRule (CPattern* cp1, CPattern* cp2)
// Parameters: (CPattern*, CPattern*)
//
//  Constructor for rule representation. 
//  Takes two arguments that are patterns of
//  lhs and rhs, respectively. 
//         								    
//////////////////////////////////////////////////////////////////////////////

CRule::CRule (CPattern* cp1, CPattern* cp2)
{
  _lhs = cp1;
  _rhs = cp2;
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: ~CRule 
// Parameters: (void)
//
// Destructor
//         								    
//////////////////////////////////////////////////////////////////////////////

CRule::~CRule(void)
{
  if(_lhs) { delete _lhs; _lhs = NULL; }
  if(_rhs) { delete _rhs; _rhs = NULL; }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: Exec
// Parameters: (CState*)
//
//  Executes the given statement.
//  Takes the state of the program before the statement
//  is executed, and then returns the state of the 
//  program after the statement is executed.
//         								    
//////////////////////////////////////////////////////////////////////////////

CState* CRule::Exec(CState* s)
{

  if(_debug_)
	cout << "-- CRule Begins" << endl;

  CState* resultState;
  resultState = _lhs->match(s);

  if(s->ISuccess()) resultState = _rhs->build(resultState);

  if(_debug_)
	cout << "-- CRule Ends" << endl;

  return resultState;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: InvExec
// Parameters: (CState*)
//
//  Executes the given statement "in reverse".
//  Takes the state of the program before the statement
//  is executed, and then returns the state of the 
//  program after the statement is executed.
//         								    
//////////////////////////////////////////////////////////////////////////////

CState* CRule::InvExec(CState* s)
{
  if(_debug_)
	cout << "-- CRule (Inv) Begins" << endl;

  CState* resultState;
  resultState = _rhs->match(s);
  
  if(s->ISuccess()) resultState = _lhs->build(resultState);
  
  if(_debug_)
	cout << "-- CRule (Inv) Ends" << endl;

  return resultState;
}
