/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Mitch Cherniack (mfc@cs.brown.edu)
//       Joon S. Lee (jl@cs.brown.edu)
// FILE: KOFJoin.C
// PROJ: COKO Compiler, KOLA Parser
// DATE: 5/97
//
//////////////////////////////////////////////////////////////////////////////

#include "KOFJoin.H"

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  FJoin
// Parameters:	(Predicate* eq, Function *f)							  
//
//  Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

FJoin::FJoin (Predicate *eq, Function *f, TFunc * ftype)
                       : Function (ftype, FUN_JOIN,"JOIN")
{
  one = eq;
  two = f;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  FJoin
// Parameters:	(FJoin& other)							  
//
//  Copy Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

FJoin::FJoin (const FJoin& other) : Function ((Function &) other)
{
  one = (other . One ()) -> Copy ();
  two = (other . Two ()) -> Copy ();
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~FJoin
// Parameters:	()							  
//
//  Destructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

FJoin::~FJoin ()
{
  delete one;
  delete two;
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Copy ()
// Parameters:	()							  
//
//  Virtual Copy Constructor 
//         								    
//////////////////////////////////////////////////////////////////////////////

Function *FJoin::Copy () const
{
  return new FJoin (*this);
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  deepeq
// Parameters:  ()
//
// Value-based equality.
// this is computed recursively.
// Return true if two trees have identical structures, otherwise return false.
// No unification occurs here.
//
//////////////////////////////////////////////////////////////////////////////

int FJoin::deepeq (NonType * other) 
{
  NameUnion* name_ = other->Myname();
  if ((name_->tag == FNAME ) && (name_->fn == FUN_JOIN))
  {
    delete name_;
    return ((one->deepeq(((FJoin*)other)->One())) 
         && (two->deepeq(((FJoin*)other)->Two())));
  }
  else
    {
      delete name_;
      return F;
    }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Myname
// Parameters:	()							  
//
//  Returns a union type, NameUnion. 
//  Since NonType has 5 different sub-classes, Function, Predicate, Object, 
//  Bool and Index, and those 4 different sub-classes use different 
//  enum type for their own sub-classes' names, Myname methods needs 
//  a union type of 5 different enum types.
//  Since Myname is defined virtually in KONonType, one can just call this 
//  methods with NonType* to identify the specific instance of Kola object.
//         								    
//////////////////////////////////////////////////////////////////////////////
 
NameUnion* FJoin::Myname()
{
	NameUnion* a = new NameUnion(FNAME, FUN_JOIN);
	return a;
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  nextChild
// Parameters:  ()
//
// Returns "currChild++"th child.
// When currChild is equal to the number of children
// (when current child is the last child), it returns NULL.
//
//////////////////////////////////////////////////////////////////////////////

NonType* FJoin::nextChild()
{
  if (currChild == 0)
	{
	  currChild++;
	  return one;
	}
  else if (currChild == 1)
	{
	  currChild++;
	  return two;
	}
  else return NULL;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  reduce
// Parameters:	()							  
//
/////////////////////////////////////////////////////////////////////////////

// OResult* FJoin::reduce (OResult* a)
// {
//   if (! (a -> isPair ())) {
//     char *or = one -> representation ();
//     char *tr = two -> representation ();
//     char *r = new char [strlen (or) + strlen (tr) + strlen ("JOIN (, )") + 1];
//     sprintf (r, "JOIN (%s, %s)", or, tr);
//     OResult *res = new OResBB (r, a -> Copy ());
//     delete [] or;
//     delete [] tr;
//     delete [] r;
//     delete a;
//     return res;
    
//   } else {
//     OResVAR *v1 = new OResVAR ();
//     OResVAR *v2 = new OResVAR ();
//     OResPAIR *a2 = ((OResPAIR *) a);
  
//     OResSET *res = 
//       new OResSET
//       (two -> reduce (new OResPAIR (v1 -> Copy (), v2 -> Copy ())),
//        new BResCOMP
//        (one -> reduce (new OResPAIR (v1 -> Copy (), v2 -> Copy ())),
// 	SAND,
// 	new BResCOMP
// 	(new BResINF (v1 -> Copy (), IN, a2 -> One () -> Copy ()),
// 	 SAND,
// 	 new BResINF (v2 -> Copy (), IN, a2 -> Two () -> Copy ()))));

//     delete v1;
//     delete v2;
//     delete a;
//     return res;
//   }
// }


// ////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  representation
// Parameters:	()							  
//
//  "FFO: FJoin (" + one -> representation () + ")"
//         								    
//////////////////////////////////////////////////////////////////////////////

char* FJoin::representation ()
{
  char *or = one -> representation ();
  char *tr = two -> representation ();
  char *res = new char [strlen (or) + strlen (tr) + strlen ("JOIN (, )") + 1];
  sprintf (res, "JOIN (%s, %s)", or, tr);
  delete [] or;    
  delete [] tr;    
  return res;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transform
// Parameters:	(CState*)							  
//
//  Return coresponding Nontype object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//
//////////////////////////////////////////////////////////////////////////////

NonType* FJoin::transform(CState* s)
{
  Predicate* first = (Predicate*) One() -> transform(s);
  Function* second = (Function*) Two() -> transform(s);
  return new FJoin(first,second,Type());
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  transformDeep
// Parameters:  (CState*, CIEnvStackDataType*)
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly.
//  The difference between this method and transform method is a search
//  method used in VAR classes. (Take a look at transform and transformDeep
//  methods in BVar, FVar, PVar, or OVar.)
//  Transform method searches the top frame only,
//  while TransformDeep methods searches entire stack starting from
//  a given point.
//
//////////////////////////////////////////////////////////////////////////////

NonType* FJoin::transformDeep(CState* s, CIEnvStackDataType* startingPt)
{
  Predicate* first = (Predicate*) One() -> transformDeep(s, startingPt);
  Function* second = (Function*) Two() -> transformDeep(s, startingPt);
  return new FJoin(first,second,Type());
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: compares
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for rule matching only.
// Rule matching is different from statement matching.
// Duplicated variable entry is allowed here.
//
// For example, foo: g o g --> g o g is a legal rule.
//                  -------   -------
//                    |         |_\
//                    |_\         / building part
//                      / matching part
//
// Assume that we apply this rule to a kola tree of the form "ID o ID."
// When unification occurs in a matching part, first we store first g and ID pair
// into the env stack. Then the next time we face another g (second one),
// we search the stack for the matching pair with variable name g and simply check
// if the matching pair in the stack has same structure as the matching pair we have.
// In this case, second pair is (g, ID) as well and the rule gets fired successfully.
//
// Now assume that we apply the same rule to a kola tree of the form "ID o (ID o ID)."
// In this case, the rule fails since the second pair is (g, ID o ID),
// while the first pair is (g, ID).
//
// Note that even in a case of failure, the rule itself is legal.
//
// "Duplication of a variable is legal in a rule section."
//
//////////////////////////////////////////////////////////////////////////////

int FJoin::compares(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();
  
  if ((tmp_name->tag == FNAME) && (tmp_name->fn == FUN_JOIN))
	{
	  // do it for first child.
	  if (one->compares(((FJoin*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one))
			// child (one) is a variable.
			{
			  if ((a->IFind(((PVar*)one)->Name())) == NULL)
				{
				  // child is a new variable.
				  a->IInsert(((PVar*)one)->Name(), new CIdent((FJoin*)nt,((FJoin*)nt)->One(),1));
				  // Push entry into the stack.
				}
			  else 
				{
				  // Child is a variable but not new one.
				  // In this case, I need to make sure that what's
				  // already in the stack and what I have now is the same.
				  if (!(((a->IFind(((PVar*)one)->Name()))->Child_Look_Up())->deepeq(((FJoin*)nt)->One())))
					return_val++;
				  // match failed.
				  // else do none.
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++;
	  // compare failed
	  
	  // do it for second child.
	  if (two->compares(((FJoin*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((FVar*)two)->Name())) == NULL)
				{
				  a->IInsert(((FVar*)two)->Name(), new CIdent((FJoin*)nt, ((FJoin*)nt)->Two(), 2));
				}
			  else 
				{
				  if (!(((a->IFind(((FVar*)two)->Name()))->Child_Look_Up())->deepeq(((FJoin*)nt)->Two())))
					return_val++;
				}
			}
		}
	  else return_val++;
	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0.
  // In this case, return false(0).
  // Otherwise, return true(1).
  
  delete tmp_name;
 
  if (return_val)
	return F;
  else
	return T;
  
} 

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: comparesSTM
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for statement matching only.
// Statement matching is different from rule matching.
// No duplicated variable entry is allowed here.
//
// For example,
//             match f ! _O,
//                   f =  g o g  is a illegal statement (equation).
//                       -------
//                         |
//                         |_\
//                           / matching part
//
// Assume that we have a query tree of the form "ID o ID ! 3."
// Variable f gets bounded to "ID o ID" and we have to match "ID o ID" to "g o g."
// Matching "ID o ID" to "g o g" is perfectly legal in a rule section.
// However, it is illegal in statement section of the program.
// Remember that every variable on the right hand side of an equation must be a fresh one.
// However, after we match first g to ID, the second g is no longer fresh variable.
//
// "No duplication of a variable is legal in a statement section."
//
//////////////////////////////////////////////////////////////////////////////

int FJoin::comparesSTM(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();
  
  if ((tmp_name->tag == FNAME) && (tmp_name->fn == FUN_JOIN))
	{
	  // do it for first child.
	  if (one->comparesSTM(((FJoin*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one))
			// child (one) is a variable.
			{
			  if ((a->IFind(((PVar*)one)->Name())) == NULL)
				{
				  // child is a new variable.
				  a->IInsert(((PVar*)one)->Name(), new CIdent((FJoin*)nt,((FJoin*)nt)->One(),1));
				  // Push entry into the stack.
				}
			  else
				{
				  return_val++;
				  cout << "Error : Variable " << ((PVar*)one)->Name() << " is used twice. " << endl;
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++;
	  // compare failed
	  
	  // do it for second child.
	  if (two->comparesSTM(((FJoin*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((FVar*)two)->Name())) == NULL)
				{
				  a->IInsert(((FVar*)two)->Name(), new CIdent((FJoin*)nt, ((FJoin*)nt)->Two(), 2));
				}
			  else 
				{
				  return_val++;
				  cout << "Error : Variable " << ((FVar*)two)->Name() << " is used twice. " << endl;
				}
			}
		}
	  else return_val++;
	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0.
  // In this case, return false(0).
  // Otherwise, return true(1).
  
  delete tmp_name;
 
  if (return_val)
	return F;
  else
	return T;

} 

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  setDEPnInvDEP
// Parameters:  (char*, CState*)
//
// set dependency list (ancestor list) and inverse dependency list (descendant list)
//
//////////////////////////////////////////////////////////////////////////////

void FJoin::setDEPnInvDEP(char* vn, CState* s)
{
  one->setDEPnInvDEP(vn,s);
  two->setDEPnInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  undoSetInvDEP
// Parameters:  (char*, CState*)
//
// undo changes made on descendant list (inverse dependency list).
// undo for dependency list (ancestor list) is not necessary, since
// any changes on dependency list is undone by poping up
// top frame from the environment stack.
//
//////////////////////////////////////////////////////////////////////////////

void FJoin::undoSetInvDEP(char* vn, CState* s)
{
  one->undoSetInvDEP(vn,s);
  two->undoSetInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  replaceChild
// Parameters:	(NonType*, int)
//
// Replace a child of given tree.  
// which_child specifies which child to replace.
// For example, when which_child is one, we replace the first child of 
// given tree.
// 
//////////////////////////////////////////////////////////////////////////////

void FJoin::replaceChild(NonType* child, int which_child)
{
  if ( which_child == 1 )
	SetOne(child);
  else if ( which_child == 2 )
	SetTwo(child);
}

#ifdef _OSTORE_
//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: invoke
//  Parameters:(koObject*)
//
//  Implements the join function former.
//
//////////////////////////////////////////////////////////////////////////////
koObject* FJoin::invoke(koObject* x)
{
  void *handle = dlopen(fjoinOpt, RTLD_LAZY);
  void *symbol = dlsym(handle, "callOpt");

  NonType *result = ((callOptPtr)symbol)(new OInvoke(this,x));
  dlclose(handle);
  return (koObject*)result;
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: exec
//  Parameters:(OSObject*)
//
//  Implements the join function former.
//
//////////////////////////////////////////////////////////////////////////////

OSObject* FJoin::exec(OSObject* x)
{
  Iterator *i = (Iterator*)((OSPair*)x)->One();
  Iterator *j = (Iterator*)((OSPair*)x)->Two();

  if (strcmp(One()->Name(),"K") == 0) {
    BPrimitive *b = ((BPrimitive*)((PConst*)One())->One());
    if (b->Value())
      return new MapJoinSieve(Two(),i,j);
  }

  if (strcmp(Two()->Name(),"ID") == 0)
    return new FilterJoinSieve(One(),i,j);

  return new JoinSieve(One(),Two(),i,j);
}

#endif // _OSTORE_
  
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  prolog
// Parameters:	()
//
// return string representation in prolog term
//
//////////////////////////////////////////////////////////////////////////////

char* FJoin::prolog ()
{
  char *or = one -> prolog ();
  char *tr = two -> prolog ();
  char *res = new char [strlen (or) + strlen (tr) + strlen ("fjoin(, )") + 1];
  sprintf (res, "fjoin(%s, %s)", or, tr);
  delete [] or;    
  delete [] tr;    
  return res;
}

#ifdef _PROLOG_
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  sp_prolog
// Parameters:	()
//
// Since C/C++ functions only have indirect access to Prolog terms 
// via SP_term_refs, we need to convert kola tree to SP_term_refs.
// Converted kola tree will get passed to prolog runtime system.
//
// See the Sicstus prolog version 3#6 manual for more detailed info on
// the usage of SP_term_refs and other prolog supplied methods.
//
//////////////////////////////////////////////////////////////////////////////

SP_term_ref FJoin::sp_prolog ()
{
  SP_term_ref or = one -> sp_prolog ();
  SP_term_ref tr = two -> sp_prolog ();
  SP_term_ref res = SP_new_term_ref();
  SP_cons_functor(res, SP_atom_from_string("fjoin"),2,or,tr);
  return res;
}
#endif // _PROLOG_

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  inv
// Parameters:	(koObject *o)
//
// Return n-ary join if appropriate
//
/////////////////////////////////////////////////////////////////////////////

koObject* FJoin::inv (koObject *o)
{
  NameUnion *nu1, *nu2f, *nu3f, *nu3p, *nub4;
  koObject *o1, *o2;
  Function *f2, *f3;
  Predicate *p3;
  koBool *b4;

  // Given this == join (p, f) and o == [join (K (true), id) ! [A, B], C]...

  o1 = o -> asPair () -> One ();
  // o1 == join (K (true), id) ! [A, B]
  nu1 = o1 -> Myname ();
  if (nu1->on == OBJ_INVOKE) {
	delete nu1; // nu1

	f2 = ((OInvoke *) o1) -> One ();
	// f2 == join (K (true), id)
    nu2f = f2 -> Myname ();

	if (nu2f->fn == FUN_JOIN) {
	  delete nu2f; // nu2

	  p3 = ((FJoin *) f2) -> One ();
	  // p3 == K (true)
	  f3 = ((FJoin *) f2) -> Two ();
	  // f3 == ID

	  nu3f = f3 -> Myname ();
	  if (nu3f->fn == FUN_ID) {
		delete nu3f; // nu3f

		nu3p = p3 -> Myname ();
		if (nu3p->pn == PRED_CONST) {
		  delete nu3p; // nu3p
		  b4 = ((PConst *) p3) -> One ();
		  // b4 == TRUE
		  nub4 = b4 -> Myname ();
		  if ((nub4->bn == BOOL_PRIM) && (((BPrimitive *) b4) -> Value () == T)) {
			delete nub4;
			return new OInvoke (Copy (), new OPair (((OInvoke *) o1) -> Two () -> Copy (), o -> asPair () -> Two () -> Copy ()));
		  } else delete nub4;
		} else delete nu3p;
	  } else delete nu3f;
	} else delete nu2f;
  } else delete nu1;

  // Otherwise, retain binary join
  return new OInvoke (Copy (), o -> Copy ());

}

/////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  sql
// Parameters:	(koObject *o)
//
// Return the SQL string equivalent of invoking this on o
// 
/////////////////////////////////////////////////////////////////////////////

char* FJoin::sql (koObject *o)
{
  NameUnion *nu;
  koObject *o1;

  nu = o -> Myname ();
  if (nu->on == OBJ_PAIR) {

	delete nu;
	o1 = o -> asPair () -> One ();
	nu = o1 -> Myname ();

	if (nu->on == OBJ_PAIR) {
	  OPair *op = new OPair (o1 -> asPair () -> Vars (), new OVar (0));
	  delete nu;

	  char *sel = (two -> inv (op)) -> sql ();
	  char *fro = o -> asPair () -> asargssql (op);
	  char *whe;

	  int flag = T;
	  NameUnion *nu = one -> Myname ();
	  if (nu -> pn == PRED_CONST) {
		delete nu;
		koBool *b = ((PConst *) one) -> One ();
		nu = b -> Myname ();
		if ((nu -> bn == BOOL_PRIM) && (((BPrimitive *) b) -> Value () == T))
		  flag = F;
	  } else delete nu;

	  if (flag)
		whe = (one -> inv (op)) -> sql ();
	  else {
		whe = new char [1];
		sprintf (whe, "");
	  };
	 
 
	  if (strncmp (fro, "ERROR", 5) == 0) {
		delete [] sel;
		delete [] whe;
		delete op;            
		return fro;
	  } else if (strncmp (sel, "ERROR", 5) == 0) {
		delete [] fro;
		delete [] whe;
		delete op;           
		return sel;
	  } else if (strncmp (whe, "ERROR", 5) == 0) {
		delete [] fro;
		delete [] sel;
		delete op;           
		return whe;
	  } else if (strcmp (whe, "") == 0) {
		char *res = new char [strlen ("(select  from )") + strlen (sel) + strlen (fro) + 1];
		sprintf (res, "(select %s from %s)", sel, fro);
		delete [] sel;
		delete [] fro;
		delete [] whe;
		delete op;           // Deleting op deletes x and y
		return res;
	  } else {
		char *res = new char [strlen ("(select  from  where )") + strlen (sel) + strlen (fro) + strlen (whe) + 1];
		sprintf (res, "(select %s from %s where %s)", sel, fro, whe);
		delete [] sel;
		delete [] fro;
		delete [] whe;
		delete op;           // Deleting op deletes x and y
		return res;
	  } 	
	} else {

	  delete nu;
	  koObject *o1 = o -> asPair () -> One ();
	  koObject *o2 = o -> asPair () -> Two ();

	  OVar *x   = new OVar (0); 
	  OVar *y   = new OVar (0); 
	  OPair *op = new OPair (x, y);

	  char *sel = (two -> inv (op)) -> sql ();
	  char *fro1 = o1 -> sql ();
	  char *fro2 = o2 -> sql ();
	  // New stuff:  Don't print out a trivial where clause
	  char *whe;
	  int flag = T;
	  NameUnion *nu = one -> Myname ();
	  if (nu -> pn == PRED_CONST) {
		delete nu;
		koBool *b = ((PConst *) one) -> One ();
		nu = b -> Myname ();
		if ((nu -> bn == BOOL_PRIM) && (((BPrimitive *) b) -> Value () == T))
		  flag = F;
	  };
	  delete nu;
	  
	  if (flag)
		whe = (one -> inv (op)) -> sql ();
	  else {
		whe = new char [1];
		sprintf (whe, "");
	  };
	  // End of New stuff
 
	  if (strncmp (fro1, "ERROR", 5) == 0) {
		delete [] sel;
		delete [] fro2;
		delete [] whe;
		delete op;            // Deleting op deletes x and y
		return fro1;
	  } else if (strncmp (fro2, "ERROR", 5) == 0) {
		delete [] sel;
		delete [] fro1;
		delete [] whe;
		delete op;            // Deleting op deletes x and y
		return fro2;
	  } else if (strncmp (sel, "ERROR", 5) == 0) {
		delete [] fro1;
		delete [] fro2;
		delete [] whe;
		delete op;           // Deleting op deletes x and y
		return sel;
	  } else if (strncmp (whe, "ERROR", 5) == 0) {
		delete [] fro1;
		delete [] fro2;
		delete [] sel;
		delete op;           // Deleting op deletes x and y
		return whe;
	  } else if (strcmp (whe, "") == 0) {
		char *res = new char [strlen ("(select  from ,   )") + strlen (sel) + strlen (x -> Name ()) + strlen (fro1) + 
							  strlen (y -> Name ()) + strlen (fro2) + 1];
		sprintf (res, "(select %s from %s %s, %s %s)", sel, fro1, x -> Name (), fro2, y -> Name ());
		delete [] sel;
		delete [] fro1;
		delete [] fro2;
		delete [] whe;
		delete op;           // Deleting op deletes x and y
		return res;
	  } else {
		char *res = new char [strlen ("(select  from  ,   where )") + strlen (sel) + strlen (x -> Name ()) + strlen (fro1) + 
							 + strlen (y -> Name ()) + strlen (fro2) + strlen (whe) + 1];
		sprintf (res, "(select %s from %s %s, %s %s where %s)", sel, fro1, x -> Name (), fro2, y -> Name (), whe);
		delete [] sel;
		delete [] fro1;
		delete [] fro2;
		delete [] whe;
		delete op;           // Deleting op deletes x and y
		return res;
	  } 	
	} 
  } else {
	delete nu;
	cerr << "ERROR: Calling JOIN on non-Pair Object" << endl;
	return NULL;
  }
}


/////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  oql
// Parameters:	(koObject *o)
//
// Return the OQL string equivalent of invoking this on o
// 
/////////////////////////////////////////////////////////////////////////////

char* FJoin::oql (koObject *o)
{
  NameUnion *nu;
  koObject *o1;

  nu = o -> Myname ();
  if (nu->on == OBJ_PAIR) {

	delete nu;
	o1 = o -> asPair () -> One ();
	nu = o1 -> Myname ();

	if (nu->on == OBJ_PAIR) {
	  OPair *op = new OPair (o1 -> asPair () -> Vars (), new OVar (0));
	  delete nu;

	  char *sel = (two -> inv (op)) -> oql ();
	  char *fro = o -> asPair () -> asargssql (op);
	  char *whe;

	  int flag = T;
	  NameUnion *nu = one -> Myname ();
	  if (nu -> pn == PRED_CONST) {
		delete nu;
		koBool *b = ((PConst *) one) -> One ();
		nu = b -> Myname ();
		if ((nu -> bn == BOOL_PRIM) && (((BPrimitive *) b) -> Value () == T))
		  flag = F;
	  } else delete nu;

	  if (flag)
		whe = (one -> inv (op)) -> oql ();
	  else {
		whe = new char [1];
		sprintf (whe, "");
	  };
	 
 
	  if (strncmp (fro, "ERROR", 5) == 0) {
		delete [] sel;
		delete [] whe;
		delete op;            
		return fro;
	  } else if (strncmp (sel, "ERROR", 5) == 0) {
		delete [] fro;
		delete [] whe;
		delete op;           
		return sel;
	  } else if (strncmp (whe, "ERROR", 5) == 0) {
		delete [] fro;
		delete [] sel;
		delete op;           
		return whe;
	  } else if (strcmp (whe, "") == 0) {
		char *res = new char [strlen ("(select  from )") + strlen (sel) + strlen (fro) + 1];
		sprintf (res, "(select %s from %s)", sel, fro);
		delete [] sel;
		delete [] fro;
		delete [] whe;
		delete op;           // Deleting op deletes x and y
		return res;
	  } else {
		char *res = new char [strlen ("(select  from  where )") + strlen (sel) + strlen (fro) + strlen (whe) + 1];
		sprintf (res, "(select %s from %s where %s)", sel, fro, whe);
		delete [] sel;
		delete [] fro;
		delete [] whe;
		delete op;           // Deleting op deletes x and y
		return res;
	  } 	
	} else {
	  delete nu;
	  koObject *o1 = o -> asPair () -> One ();
	  koObject *o2 = o -> asPair () -> Two ();

	  OVar *x   = new OVar (0); 
	  OVar *y   = new OVar (0); 
	  OPair *op = new OPair (x, y);

	  char *sel = (two -> inv (op)) -> oql ();
	  char *fro1 = o1 -> oql ();
	  char *fro2 = o2 -> oql ();
	  // New stuff:  Don't print out a trivial where clause
	  char *whe;
	  int flag = T;
	  NameUnion *nu = one -> Myname ();
	  if (nu -> pn == PRED_CONST) {
		delete nu;
		koBool *b = ((PConst *) one) -> One ();
		nu = b -> Myname ();
		if ((nu -> bn == BOOL_PRIM) && (((BPrimitive *) b) -> Value () == T))
		  flag = F;
	  };
	  delete nu;
	  
	  if (flag)
		whe = (one -> inv (op)) -> oql ();
	  else {
		whe = new char [1];
		sprintf (whe, "");
	  };
	  // End of New stuff
 
	  if (strncmp (fro1, "ERROR", 5) == 0) {
		delete [] sel;
		delete [] fro2;
		delete [] whe;
		delete op;            // Deleting op deletes x and y
		return fro1;
	  } else if (strncmp (fro2, "ERROR", 5) == 0) {
		delete [] sel;
		delete [] fro1;
		delete [] whe;
		delete op;            // Deleting op deletes x and y
		return fro2;
	  } else if (strncmp (sel, "ERROR", 5) == 0) {
		delete [] fro1;
		delete [] fro2;
		delete [] whe;
		delete op;           // Deleting op deletes x and y
		return sel;
	  } else if (strncmp (whe, "ERROR", 5) == 0) {
		delete [] fro1;
		delete [] fro2;
		delete [] sel;
		delete op;           // Deleting op deletes x and y
		return whe;
	  } else if (strcmp (whe, "") == 0) {
		char *res = new char [strlen ("(select  from ,   )") + strlen (sel) + strlen (x -> Name ()) + strlen (fro1) + 
							  strlen (y -> Name ()) + strlen (fro2) + 1];
		sprintf (res, "(select %s from %s %s, %s %s)", sel, fro1, x -> Name (), fro2, y -> Name ());
		delete [] sel;
		delete [] fro1;
		delete [] fro2;
		delete [] whe;
		delete op;           // Deleting op deletes x and y
		return res;
	  } else {
		char *res = new char [strlen ("(select  from  ,   where )") + strlen (sel) + strlen (x -> Name ()) + strlen (fro1) + 
							 + strlen (y -> Name ()) + strlen (fro2) + strlen (whe) + 1];
		sprintf (res, "(select %s from %s %s, %s %s where %s)", sel, fro1, x -> Name (), fro2, y -> Name (), whe);
		delete [] sel;
		delete [] fro1;
		delete [] fro2;
		delete [] whe;
		delete op;           // Deleting op deletes x and y
		return res;
	  } 	
	} 
  } else {
	delete nu;
	cerr << "ERROR: Calling JOIN on non-Pair Object" << endl;
	return NULL;
  }
}
