/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Joon S. Lee (jl@cs.brown.edu)
//
// FILE: KOFVar.C
// PROJ: COKO Compiler
// DATE: 5/97
//
//////////////////////////////////////////////////////////////////////////////

#include "KOFVar.H"

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  FVar
// Parameters:	(char* n)							  
//
//  Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

FVar::FVar  (char* s, TFunc * ftype) : Function (ftype, FUN_VAR, s)
{
  if (s == NULL) val = NULL;
  else {
    val = new char [strlen(s) + 1];
    sprintf (val, "%s", s);
  }
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  FVar
// Parameters:	(FVar& other)							  
//
//  Copy Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

FVar::FVar (const FVar& other) : Function ((Function &) other)
{
  if (other.val == NULL) val = NULL;
  else {
    val = new char [strlen(other.val) + 1];
    sprintf (val, "%s", other.val);
  }
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~FVar
// Parameters:	()							  
//
//  Destructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

FVar::~FVar ()
{
 if (val) {delete [] val; val = NULL;}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  reduce
// Parameters:	()							  
//
/////////////////////////////////////////////////////////////////////////////

// OResult* FVar::reduce (OResult* a)
// {
//     char *r = new char [strlen (val) + strlen ("(F: )") + 1];
//     sprintf (r, "(F: %s)", val);
//     OResult *res = new OResBB (r, a -> Copy ());
//     delete [] r;
//     delete a;
//     return res;
// }

// //////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  representation
// Parameters:	()							  
//
//  "FVAR: " + name
//         								    
//////////////////////////////////////////////////////////////////////////////

char* FVar::representation ()
{

    char *res = new char [strlen (val) + strlen ("(F: )") + 1];
    sprintf (res, "(F: %s)", val);

    return res;


}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Copy
// Parameters:	()							  
//
//  return copy of this as Function*
//         								    
//////////////////////////////////////////////////////////////////////////////

Function* FVar::Copy () const
{
    return new FVar (*this);
}



//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transform
// Parameters:	(CState*)							  
//
//  if there is a entry for this variable name in the stack
//  return the corresponding Nontype else return NULL
//         								    
//////////////////////////////////////////////////////////////////////////////

NonType* FVar::transform(CState* a)
{
  if ((a->IFind(val)) != NULL)
	{
	  a->SetSuccess(T);
	  return ((Function*)((a->IFind(val))->Child()));
	}
  else
	{
	  a->SetSuccess(F);
	  return NULL;
	}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transformDeep
// Parameters:	(CState*, CIEnvStackDataType*)							  
//
//  if there is a entry for this variable name in the stack
//  return the corresponding NonType else return NULL
//  Start search from given point.
//         								    
//////////////////////////////////////////////////////////////////////////////

NonType* FVar::transformDeep(CState* s,CIEnvStackDataType* startingPt)
{
  if (s->IFindFromHere(val, startingPt) != NULL)
	{
	  s->SetSuccess(T);
	  return ((Function*)((s->IFindFromHere(val, startingPt))->Child()));
	}
  else
	{
	  s->SetSuccess(F);
	  return NULL;
	}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  compares
// Parameters:	(NonType*, CState*)							  
//
//  For variables, I only need to check whether the corresponding
//  NonType Class is as same type as current NonType.
//
//////////////////////////////////////////////////////////////////////////////

int FVar::compares(NonType* nt, CState* a)
{

  // Please check: is this right?
  // (tmp_name->fn >= FUN_CHOOSE) && (tmp_name->fn <= FUN_ITER))
  NameUnion* tmp_name =  nt->Myname();
  if ((tmp_name->tag == FNAME) && (tmp_name->fn != FUN_INVALID) &&
	  (tmp_name->fn != FUN_DC) && (tmp_name->fn != FUN_VAR)) 
	{
	  delete tmp_name;
	  return T;
	}
  else
	{
	  delete tmp_name;
	  return F;
	}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Myname
// Parameters:	()							  
//
//  Returns a union type, NameUnion. 
//  Since NonType has 5 different sub-classes, Function, Predicate, Object, 
//  Bool and Index, and those 4 different sub-classes use different 
//  enum type for their own sub-classes' names, Myname methods needs 
//  a union type of 5 different enum types.
//  Since Myname is defined virtually in KONonType, one can just call this 
//  methods with NonType* to identify the specific instance of Kola object.
//         								    
//////////////////////////////////////////////////////////////////////////////

NameUnion* FVar::Myname()
{
	NameUnion* a = new NameUnion(FNAME, FUN_VAR);
	return a;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: directStore
// Parameters:	(CState*)							  
//
//  For variables, I only need to check whether the corresponding
//  Nontype Class is as same type as current NonType.
//
//////////////////////////////////////////////////////////////////////////////

CState*  FVar::directStore(CState* a)
{
  if (a->Store() == NULL)
	{
	  cout << "EOR -------- EOR ---- Store is NULL :: FVar" << endl;
	  return NULL;
	}
  
  NameUnion* tmp_name =  (a->Store())->Myname();
  
  // if ((tmp_name->tag == FNAME) && (tmp_name->fn >= FUN_CHOOSE) && (tmp_name->fn <= FUN_ITER))
  if ((tmp_name->tag == FNAME) && (tmp_name->fn != FUN_INVALID) &&
	  (tmp_name->fn != FUN_DC) && (tmp_name->fn != FUN_VAR))
	{
	  
	  if (((a->IFind(val))) == NULL)
		{
		  a->IInsert(val, new CIdent( NULL, a->Store(), 0));
		  a->SetSuccess(T);
		}
	  else if (((a->IFind(val))) != NULL)
		{
		  if (!((a->IFind(val))->Child_Look_Up())->deepeq (a->Store()))
			a->SetSuccess(F);
		  else
			a->SetSuccess(T);
		}
	}
  else a->SetSuccess(F);
  
  delete tmp_name;
  return a;
}      


//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  setInvDEP
// Parameters:  (char*, CState*)
//
// set dependency list and inverse dependency list
//
//////////////////////////////////////////////////////////////////////////////

void FVar::setDEPnInvDEP(char* vn, CState* s)
{
  if (vn == NULL)
	  s->CInsertInvDep(val);
  else
	{
	  s->IInsertInvDep(vn, val);
	  s->IInsertDep(val, vn);
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  undoSetInvDEP
// Parameters:  (char*, CState*)
//
// undo changes made on inverse dependency list (ansester list).
// undo for dependency list is not necessary, since 
// any changes on dependency list is undone by poping up
// top frame from the environment stack.
//
//////////////////////////////////////////////////////////////////////////////

void FVar::undoSetInvDEP(char* vn, CState* s)
{
  s->IDeleteInvDep(val);
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  replaceChild
// Parameters:	(NonType*, int)
//
// Replace a child of given tree.  
// which_child specifies which child to replace.
// For example, when which_child is one, we replace the first child of 
// given tree.
// 
//////////////////////////////////////////////////////////////////////////////

void FVar::replaceChild(NonType* child, int which_child)
{
}
  
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  prolog
// Parameters:	()
//
// return string representation in prolog term
//
//////////////////////////////////////////////////////////////////////////////

char* FVar::prolog ()
{
  char *res = new char [strlen (val) + strlen ("V") + 1];
  sprintf (res, "V%s", val);
  return res;
}

#ifdef _PROLOG_
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  sp_prolog
// Parameters:	()
//
// Since C/C++ functions only have indirect access to Prolog terms 
// via SP_term_refs, we need to convert kola tree to SP_term_refs.
// Converted kola tree will get passed to prolog runtime system.
//
// See the Sicstus prolog version 3#6 manual for more detailed info on
// the usage of SP_term_refs and other prolog supplied methods.
//
//////////////////////////////////////////////////////////////////////////////

SP_term_ref FVar::sp_prolog ()
{
  cout << "Error @FVar::sp_prlog" << endl;
  cout << "query input should not have variable inside error!" << endl;
  return NULL;
}
#endif // _PROLOG_

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Val
// Parameters:	()
//
// Selector - return a copy of val
//         								    
//////////////////////////////////////////////////////////////////////////////

char* FVar::Val ()
{
  char *res = new char [strlen (val) + 1];
  sprintf (res, "%s", val);
  return res;
}
