/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//                                                                          
//  NAME: Mitch Cherniack (mfc)
//        Joon S. Lee (jl)					 	    
//  FILE: KOObjList.H 
//  PROJ: COKO, KOLA
//  DATE: 5/97		                                                    
//
// SYNOPSIS
//  List of objects that denote members of a set
//  in a KOLA tree.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef KOLA_OBJLIST_HEADER
#define KOLA_OBJLIST_HEADER 

#include "KGlobals.H"
#include "KOObject.H"
#include "KOOLink.H"
#include "PCState.H"
#include "KONameUnion.H"

#include <string.h>

class ObjList
{
public:
  ObjList ();
  //  Constructor.  Construct an empty list (set)
  //    first = NULL; last = NULL;
  
  ObjList (const ObjList &other);
  //  Copy Constructor.  

  ~ObjList ();
  //  Destructor.  

  OLink* First () const {return first;}
  //  Return first link in list.
  
  OLink* Last () const {return last;}
  //  Return last link in list.
  
  ObjList* insert (koObject* o);
  //  insert object into set:

  int listeq(ObjList* otherList);    
  // Compare two list.
  // every element in the first list should be in the second list
  // and vice verse
  // added for deepeq() in KOOSet class.
  // Added by Joon Lee 1996 7 5.
  // Not Yet Tested in real situation.
  // It might have some logical bugs.
  // I will test this methods later.

  //  OResChain* reduce ();
  //  Return 
  //    new OResChain (first, last)
	  
  char* prolog ();
  // return string representation in prolog term

  char* representation ();

  char *sql ();
  // Return the SQL string equivalent of this object

  char *oql ();
  // Return the OQL string equivalent of this object

  ObjList *objlist ();
  // Symbolic simplification of all objects in list.

#ifdef _PROLOG_
  SP_term_ref sp_prolog ();
    // convert kola tree into prolog term and return it.
#endif // _PROLOG_ 

protected:
  OLink* first;
  //  What is the first object in this list?  (NULL
  //  if the list (and hence set) is empty)
	
  OLink* last;
  //  What is the last object in this list (set)?  NULL
  //  if the list (and hence set) is empty.
  
};

# endif



