/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Mitch Cherniack (mfc@cs.brown.edu)
//       Joon S. Lee (jl@cs.brown.edu)
// FILE: KOPEqual.C
// PROJ: COKO Compiler, KOLA Parser
// DATE: 5/97
//
//////////////////////////////////////////////////////////////////////////////

#include "KOPEqual.H"

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  PEqual
// Parameters:	()							  
//
//  Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

PEqual::PEqual (TPred * t)  : Predicate (t, PRED_EQUAL,"=")
{    
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  PEqual
// Parameters:	(PEqual& other)							  
//
//  Copy Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

PEqual::PEqual (const PEqual& other) : Predicate ((Predicate &) other)
{
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~PEqual
// Parameters:	()							  
//
//  Destructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

PEqual::~PEqual ()
{
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Copy
// Parameters:	()							  
//
//  return copy of this as Predicate*
//         								    
//////////////////////////////////////////////////////////////////////////////

Predicate* PEqual::Copy () const
{
   return new PEqual (*this);
}


//////////////////////////////////////////////////////////////////////////////
//
// Function Name:deepeq
// Parameters:(Nontype*)
//
// Value-based equality.
// this is computed recursively
//
//////////////////////////////////////////////////////////////////////////////

int PEqual::deepeq (NonType * other) 
{
  NameUnion* name_ = other->Myname();
  if ((name_->tag == PNAME ) && (name_->pn == PRED_EQUAL))
  {
    delete name_;
    return T;
  }
  else
  {
    delete name_;
    return F;
  }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Myname
// Parameters:	()							  
//
//  Returns a union type, NameUnion. 
//  Since NonType has 5 different sub-classes, Function, Predicate, Object, 
//  Bool and Index, and those 4 different sub-classes use different 
//  enum type for their own sub-classes' names, Myname methods needs 
//  a union type of 5 different enum types.
//  Since Myname is defined virtually in KONonType, one can just call this 
//  methods with NonType* to identify the specific instance of Kola object.
//         								    
//////////////////////////////////////////////////////////////////////////////

NameUnion* PEqual::Myname()
{
	NameUnion* a = new NameUnion(PNAME, PRED_EQUAL);
	return a;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  representation
// Parameters:	()							 
//
//  "PPRIM: EQUAL"
//         								    
//////////////////////////////////////////////////////////////////////////////

char* PEqual::representation ()
{
  char *res = new char [strlen ("=") + 1];
  sprintf (res, "%s", "=");
  return res;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: representation 
// Parameters:	()							  
//  return new BResINF (a -> fst -> reduce (), "=", 
//  		    a -> snd -> reduce ())
//
/////////////////////////////////////////////////////////////////////////////

// BResult* PEqual::reduce (OResult* a)
// {
//   if (a -> isPair ()) {
//     OResPAIR *a2 = ((OResPAIR *) a);
//     BResINF *res =  new BResINF
//       (a2 -> One () -> Copy (), "=", a2 -> Two () -> Copy ());
//     delete a;
//     return res;
//   } else {
//     BResBB *res = new BResBB ("EQUAL", a -> Copy ());
//     delete a;
//     return res;
//   }
// }

// //////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transform
// Parameters:	(CState*)							  
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//
//////////////////////////////////////////////////////////////////////////////

NonType* PEqual::transform(CState* s)
{
  
    return new PEqual(Type());
    
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: compares
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for rule matching only.
// Rule matching is different from statement matching.
// Duplicated variable entry is allowed here. 
//
//////////////////////////////////////////////////////////////////////////////

int PEqual::compares(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();

  if ((tmp_name->tag == PNAME) && (tmp_name->pn == PRED_EQUAL))
	{
	  delete tmp_name;
	  return T;
	}
  else 
	{
	  delete tmp_name;
	  return F;
	}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  replaceChild
// Parameters:	(NonType*, int)
//
// Replace a child of given tree.  
// which_child specifies which child to replace.
// For example, when which_child is one, we replace the first child of 
// given tree.
// 
//////////////////////////////////////////////////////////////////////////////

void PEqual::replaceChild(NonType* child, int which_child)
{
}

#ifdef _OSTORE_
//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: invoke
//  Parameters:(koObject*)
//
//  Implements the equal predicate object.
//
//////////////////////////////////////////////////////////////////////////////

koBool* PEqual::invoke(koObject* x)
{
  return new BInvoke(this,x);
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: exec
//  Parameters:(OSObject*)
//
//  Implements the equal predicate object.
//
//////////////////////////////////////////////////////////////////////////////

OSBool* PEqual::exec(OSObject* x)
{
  koObject *i = (((OSPair*)x)->One())->kola();
  koObject *j = (((OSPair*)x)->Two())->kola();
  if (i->equal(j))
    return new OSBPrimitive(1);
  else
    return new OSBPrimitive(0);
}

#endif // _OSTORE_
  
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  prolog
// Parameters:	()
//
// return string representation in prolog term
//
//////////////////////////////////////////////////////////////////////////////

char* PEqual::prolog ()
{
  char *res = new char [strlen ("pequal") + 1];
  sprintf (res, "pequal");
  return res;
}

#ifdef _PROLOG_
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  sp_prolog
// Parameters:	()
//
// Since C/C++ functions only have indirect access to Prolog terms 
// via SP_term_refs, we need to convert kola tree to SP_term_refs.
// Converted kola tree will get passed to prolog runtime system.
//
// See the Sicstus prolog version 3#6 manual for more detailed info on
// the usage of SP_term_refs and other prolog supplied methods.
//
//////////////////////////////////////////////////////////////////////////////

SP_term_ref PEqual::sp_prolog ()
{
  SP_term_ref res = SP_new_term_ref();
  SP_put_string(res,"pequal");
  return res;
}
#endif // _PROLOG_

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  sql
// Parameters:	(koObject *o)
//
// Return the SQL string equivalent of invoking this on o
//
/////////////////////////////////////////////////////////////////////////////

char* PEqual::sql (koObject *o)
{    
    NameUnion *nu;
    nu = o -> Myname ();
    if (nu -> on == OBJ_PAIR) {
	  delete nu;
	  char *arg1 = o -> asPair () -> One () -> sql ();
	  char *arg2 = o -> asPair () -> Two () -> sql ();
	  if (strncmp (arg1, "ERROR", 5) == 0) {
		delete [] arg2;
		return arg1;
	  } else if (strncmp (arg2, "ERROR", 5) == 0) {
		delete [] arg1;
		return arg2;
	  } else {
		char *res = new char [strlen ("( = )") + strlen (arg1) + strlen (arg2) + 1];
		sprintf (res, "(%s = %s)", arg1, arg2);
		delete [] arg1;
		delete [] arg2;
		return res;
	  }
	} else {
	  delete nu;
	  cerr << "ERROR: Calling EQUAL on non-Pair Object" << endl;
	  return NULL;
	}
}



//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  oql
// Parameters:	(koObject *o)
//
// Return the OQL string equivalent of invoking this on o
//
/////////////////////////////////////////////////////////////////////////////

char* PEqual::oql (koObject *o)
{    
    NameUnion *nu;
    nu = o -> Myname ();
    if (nu -> on == OBJ_PAIR) {
	  delete nu;
	  char *arg1 = o -> asPair () -> One () -> oql ();
	  char *arg2 = o -> asPair () -> Two () -> oql ();
	  if (strncmp (arg1, "ERROR", 5) == 0) {
		delete [] arg2;
		return arg1;
	  } else if (strncmp (arg2, "ERROR", 5) == 0) {
		delete [] arg1;
		return arg2;
	  } else {
		char *res = new char [strlen ("( = )") + strlen (arg1) + strlen (arg2) + 1];
		sprintf (res, "(%s = %s)", arg1, arg2);
		delete [] arg1;
		delete [] arg2;
		return res;
	  }
	} else {
	  delete nu;
	  cerr << "ERROR: Calling EQUAL on non-Pair Object" << endl;
	  return NULL;
	}
}

