/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Mitch Cherniack (mfc@cs.brown.edu)
//       Joon S. Lee (jl@cs.brown.edu)			
// FILE: KOPLex.C
// PROJ: COKO Compiler, KOLA Parcer
// DATE: 5/97
//
//////////////////////////////////////////////////////////////////////////////
 
#include "KOPLex.H"
//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  PLex
// Parameters:	(Predicate* p, Predicate *q)							  
//
//  Constructor.  p and q are predicate arguments.
//         								    
//////////////////////////////////////////////////////////////////////////////

PLex::PLex (Predicate* p1, Predicate *p2, TPred * t) : Predicate (t, PRED_LEX,"LEX")
{
    one = p1;
    two = p2;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  PLex
// Parameters:	(PLex& other)		     			  
//
//  Copy Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

PLex::PLex (const PLex& other) : Predicate ((Predicate &) other)
{
    one = other . One () -> Copy ();
    two = other . Two () -> Copy ();
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~PLex
// Parameters:	()
//
//  Destructor
//         								    
//////////////////////////////////////////////////////////////////////////////

PLex::~PLex ()
{
    delete one;
    delete two;
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Copy ()
// Parameters:	()
//
//  Return copy of *this as Predicate
//         								    
//////////////////////////////////////////////////////////////////////////////

Predicate* PLex::Copy () const
{
    return new PLex (*this);
}


//////////////////////////////////////////////////////////////////////////////
//
// Function Name:deepeq
// Parameters:(Nontype*)
//
// Value-based equality.
// this is computed recursively
//
//////////////////////////////////////////////////////////////////////////////

int PLex::deepeq (NonType * other) 
{
  NameUnion* name_ = other->Myname();
  if ((name_->tag == PNAME ) && (name_->pn == PRED_LEX))
  {
    delete name_;
    return ((one->deepeq(((PLex*)other)->One())) && (two->deepeq(((PLex*)other)->Two())));
  }
  else
    {
      delete name_;
      return F;
    }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Myname
// Parameters:	()							  
//
//  Returns a union type, NameUnion. 
//  Since NonType has 5 different sub-classes, Function, Predicate, Object, 
//  Bool and Index, and those 4 different sub-classes use different 
//  enum type for their own sub-classes' names, Myname methods needs 
//  a union type of 5 different enum types.
//  Since Myname is defined virtually in KONonType, one can just call this 
//  methods with NonType* to identify the specific instance of Kola object.
//         								    
//////////////////////////////////////////////////////////////////////////////

NameUnion* PLex::Myname()
{
	NameUnion* a = new NameUnion(PNAME, PRED_LEX);
	return a;
}

//////////////////////////////////////////////////////////////////////////////
//
//   Function Name: nextChild
//   Parameters: ()
// 
// Returns "currChild++"th child.
// When currChild is equal to the number of children
// (when current child is the last child), it returns NULL.
//
//////////////////////////////////////////////////////////////////////////////

NonType* PLex::nextChild()
{
  if (currChild == 0)
	{
	  currChild++;
	  return one;
	}
  else if (currChild == 1)
	{
	  currChild++;
	  return two;
	}
  else return NULL;
}

////////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  representation
// Parameters:	()							  
//
//  "PFO: LEX" + one -> representation () + two -> representation ()
//         								    
//////////////////////////////////////////////////////////////////////////////

char* PLex::representation ()
{

    char *or = one -> representation ();
    char *tr = two -> representation ();
    
    char *res = new char [strlen (or) + strlen (tr) + strlen ("LEX (, )") + 1];
    sprintf (res, "LEX (%s, %s)", or, tr);
    delete [] or;
    delete [] tr;
    return res;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transform
// Parameters:	(CState*)							  
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//
//////////////////////////////////////////////////////////////////////////////

NonType* PLex::transform(CState* s)
{
  Predicate* first = (Predicate*) One() -> transform(s);
  Predicate* second = (Predicate*) Two() -> transform(s);
  return new PLex(first, second,Type());
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transformDeep
// Parameters:	(CState*, CIEnvStackDataType*)							  
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//  The difference between this method and transform method is a search
//  method used in VAR classes. (Take a look at transform and transformDeep
//  methods in BVar, FVar, PVar, or OVar.
//
//////////////////////////////////////////////////////////////////////////////

NonType* PLex::transformDeep(CState* s, CIEnvStackDataType* startingPt)
{
  Predicate* first = (Predicate*) One() -> transformDeep(s, startingPt);
  Predicate* second = (Predicate*) Two() -> transformDeep(s, startingPt);
  return new PLex(first, second,Type());
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: compares
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for rule matching only.
// Rule matching is different from statement matching.
// Duplicated variable entry is allowed here. 
//
//////////////////////////////////////////////////////////////////////////////

int PLex::compares(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();
  
  if ((tmp_name->tag == PNAME) && (tmp_name->pn == PRED_LEX))
	{
	  // do it for first child.
	  if (one->compares(((PLex*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one))
			// child (one) is a variable.
			{
			  if ((a->IFind(((PVar*)one)->Name())) == NULL)
				{
				  // child is a new variable.
				  a->IInsert(((PVar*)one)->Name(), new CIdent((PLex*)nt,((PLex*)nt)->One(),1));
				  // Push entry into the stack.
				}
			  else 
				{
				  // Child is a variable but not new one.
				  // In this case, I need to make sure that what's
                                  // already in the stack and what I have now is the same.
				  if (!(((a->IFind(((PVar*)one)->Name()))->Child_Look_Up())->deepeq(((PLex*)nt)->One())))
					return_val++;
				  // match failed.
				  // else do none.
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++;
	  // compare failed
	  
	  // do it for second child.
	  if (two->compares(((PLex*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((PVar*)two)->Name())) == NULL)
				{
				  a->IInsert(((PVar*)two)->Name(), new CIdent((PLex*)nt, ((PLex*)nt)->Two(), 2));
				}
			  else 
				{
				  if (!(((a->IFind(((PVar*)two)->Name()))->Child_Look_Up())->deepeq(((PLex*)nt)->Two())))
					return_val++;
				}
			}
		}
	  else return_val++;
	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0.
  // In this case, return false(0).
  // Otherwise, return true(1).
  
  delete tmp_name;
 
  if (return_val)
	return F;
  else
	return T;

} 

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: comparesSTM
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This match method is used for matching statement.
// No duplicated variable entry is allowed.
//
//////////////////////////////////////////////////////////////////////////////

int PLex::comparesSTM(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();
  
  if ((tmp_name->tag == PNAME) && (tmp_name->pn == PRED_LEX))
	{
	  // do it for first child.
	  if (one->comparesSTM(((PLex*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one))
			// child (one) is a variable.
			{
			  if ((a->IFind(((PVar*)one)->Name())) == NULL)
				{
				  // child is a new variable.
				  a->IInsert(((PVar*)one)->Name(), new CIdent((PLex*)nt,((PLex*)nt)->One(),1));
				  // Push entry into the stack.
				}
			  else 
				{
				  return_val++;
				  cout << "Error : Variable " << ((PVar*)one)->Name() << " is used twice. " << endl;
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++;
	  // compare failed
	  
	  // do it for second child.
	  if (two->comparesSTM(((PLex*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((PVar*)two)->Name())) == NULL)
				{
				  a->IInsert(((PVar*)two)->Name(), new CIdent((PLex*)nt, ((PLex*)nt)->Two(), 2));
				}
			  else 
				{
				  return_val++;
				  cout << "Error : Variable " << ((PVar*)two)->Name() << " is used twice. " << endl;
				}
			}
		}
	  else return_val++;
	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0.
  // In this case, return false(0).
  // Otherwise, return true(1).
  
  delete tmp_name;
 
  if (return_val)
	return F;
  else
	return T;

} 

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  setDEPnInvDEP
// Parameters:  (char*, CState*)
//
// set dependency list and inverse dependency list
//
//////////////////////////////////////////////////////////////////////////////

void PLex::setDEPnInvDEP(char* vn, CState* s)
{
  one->setDEPnInvDEP(vn,s);
  two->setDEPnInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  undoSetInvDEP
// Parameters:  (char*, CState*)
//
// undo changes made on inverse dependency list.
// undo for dependency list is not necessary, since 
// any changes on dependency list is undone by poping up
// top frame from the environment stack.
//
//////////////////////////////////////////////////////////////////////////////

void PLex::undoSetInvDEP(char* vn, CState* s)
{
  one->undoSetInvDEP(vn,s);
  two->undoSetInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  replaceChild
// Parameters:	(NonType*, int)
//
// Replace a child of given tree.  
// which_child specifies which child to replace.
// For example, when which_child is one, we replace the first child of 
// given tree.
// 
//////////////////////////////////////////////////////////////////////////////

void PLex::replaceChild(NonType* child, int which_child)
{
  if ( which_child == 1 )
	SetOne(child);
  else if ( which_child == 2 )
	SetTwo(child);
}

#ifdef _OSTORE_
//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: invoke
//  Parameters:(koObject*)
//
//  Implements the predicate lexion former for koObjects.
//
//////////////////////////////////////////////////////////////////////////////

koBool* PLex::invoke(koObject* x)
{
  return new BInvoke(this,x);
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: exec
//  Parameters:(OSObject*)
//
//  Implements the predicate lexion former for OSObjects.
//
//////////////////////////////////////////////////////////////////////////////

OSBool* PLex::exec(OSObject* x)
{
  cerr << "PLex::exec being invoked when undefined." << end;
  return NULL;
  //  if (((OSBPrimitive*)(One()->exec(x)))->Val() &&
  //      ((OSBPrimitive*)(Two()->exec(x)))->Val())
  //    return new OSBPrimitive(1);
  //  else
  //    return new OSBPrimitive(0);
}
#endif // _OSTORE_

  
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  prolog
// Parameters:	()
//
// return string representation in prolog term
//
//////////////////////////////////////////////////////////////////////////////

char* PLex::prolog ()
{

  char *or = one -> prolog ();
  char *tr = two -> prolog ();
  
  char *res = new char [strlen (or) + strlen (tr) + strlen ("plex(, )") + 1];
  sprintf (res, "plex(%s, %s)", or, tr);
  delete [] or;
  delete [] tr;
  return res;
}

#ifdef _PROLOG_
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  sp_prolog
// Parameters:	()
//
// Since C/C++ functions only have indirect access to Prolog terms 
// via SP_term_refs, we need to convert kola tree to SP_term_refs.
// Converted kola tree will get passed to prolog runtime system.
//
// See the Sicstus prolog version 3#6 manual for more detailed info on
// the usage of SP_term_refs and other prolog supplied methods.
//
//////////////////////////////////////////////////////////////////////////////

SP_term_ref PLex::sp_prolog ()
{
  SP_term_ref or = one -> sp_prolog ();
  SP_term_ref tr = two -> sp_prolog ();
  SP_term_ref res = SP_new_term_ref();
  SP_cons_functor(res, SP_atom_from_string("plex"),2,or,tr);
  return res;
}
#endif // _PROLOG_

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  inv
// Parameters:	(koObject *o)
//
// Symbolic invocation of this on o.  Eventually,
// this method will replace "invoke" presently
// only defined with _OSTORE_ flag set
//
/////////////////////////////////////////////////////////////////////////////

koBool *PLex::inv (koObject *o) 
{
  NameUnion *nu, *nu1, *nu2;

  nu = o -> Myname ();
  if (nu ->on == OBJ_PAIR) {
	delete nu;
	koObject *o1 = o -> asPair () -> One ();
	koObject *o2 = o -> asPair () -> Two ();
    nu1 = o1 -> Myname ();
    nu2 = o2 -> Myname ();
	if ((nu1 -> on == OBJ_PAIR) && (nu2 -> on == OBJ_PAIR)) {
	  delete nu1;
	  delete nu2;
	  OPair *op1 = new OPair (o1 -> asPair () -> One () -> Copy (), o2 -> asPair () -> One () -> Copy ());
	  OPair *op2 = new OPair (o1 -> asPair () -> Two () -> Copy (), o2 -> asPair () -> Two () -> Copy ());

	  koBool *res = new BOr (one -> inv (op1), new BAnd (new BEqu (o1 -> asPair () -> One (), o2 -> asPair () -> One ()), 
														 two -> inv (op2)));
	  delete op1;
	  delete op2;
	  return res;
	} else {	  
	  delete nu1;
	  delete nu2;
	  cerr << "Invoking PLex on Object that is not of Pair of Pairs" << endl;
	  return NULL;
    }
  } else {
	cerr << "Invoking PLex on Object that is not of Pair of Pairs" << endl;	
    return NULL;
  }
}
