/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Mitch Cherniack (mfc@cs.brown.edu)
//       Joon S. Lee (jl@cs.brown.edu)			  
// FILE: KOPTimes.C
// PROJ: COKO Compiler, KOLA Parser
// DATE: 5/97
//
//////////////////////////////////////////////////////////////////////////////
 
#include "KOPTimes.H"
//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  PTimes
// Parameters:	(Predicate* p1, Predicate *p2)		    
//
//  Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

PTimes::PTimes (Predicate* p1, Predicate *p2, TPred * t)
                         : Predicate (t, PRED_TIMES,"x")
{
    one = p1;
    two = p2;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  PTimes
// Parameters:	(PTimes& other)				
//
//  Copy Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

PTimes::PTimes (const PTimes& other) : Predicate ((Predicate &) other)
{
    one = other . One () -> Copy ();
    two = other . Two () -> Copy ();
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~PTimes
// Parameters:	()
//
//  Destructor
//         								    
//////////////////////////////////////////////////////////////////////////////

PTimes::~PTimes ()
{
    delete one;
    delete two;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Copy ()
// Parameters:	()
//
//  Return copy of *this as Predicate
//         								    
//////////////////////////////////////////////////////////////////////////////

Predicate* PTimes::Copy () const
{

    return new PTimes (*this);
    
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:deepeq
// Parameters:(Nontype*)
//
// Value-based equality.
// this is computed recursively
//
//////////////////////////////////////////////////////////////////////////////

int PTimes::deepeq (NonType * other) 
{
  NameUnion* name_ = other->Myname();
  if ((name_->tag == PNAME ) && (name_->pn == PRED_TIMES))
  {
    delete name_;
    return ((one->deepeq(((PTimes*)other)->One())) && (two->deepeq(((PTimes*)other)->Two())));
  }
  else
    {
      delete name_;
      return F;
    }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Myname
// Parameters:	()							  
//
//  Returns a union type, NameUnion. 
//  Since NonType has 5 different sub-classes, Function, Predicate, Object, 
//  Bool and Index, and those 5 different sub-classes use different 
//  enum type for their own sub-classes' names, Myname methods needs 
//  a union type of 5 different enum types.
//  Since Myname is defined virtually in KONonType, one can just call this 
//  methods with NonType* to identify the specific instance of Kola object.
//         								    
//////////////////////////////////////////////////////////////////////////////

NameUnion* PTimes::Myname()
{
	NameUnion* a = new NameUnion(PNAME, PRED_TIMES);
	return a;
}

//////////////////////////////////////////////////////////////////////////////
//
//   Function Name: nextChild
//   Parameters: ()
// 
// Returns "currChild++"th child.
// When currChild is equal to the number of children
// (when current child is the last child), it returns NULL.
//
//////////////////////////////////////////////////////////////////////////////

NonType* PTimes::nextChild()
{
  if (currChild == 0)
	{
	  currChild++;
	  return one;
	}
  else if (currChild == 1)
	{
	  currChild++;
	  return two;
	}
  else return NULL;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  reduce
// Parameters:	(OResult* a)						
//
//  return new BResCOM
//    (p1 -> reduce (a -> fst -> reduce ()), "/\", 
//     p2 -> reduce (a -> snd -> reduce ()))
//         								    
//////////////////////////////////////////////////////////////////////////////

// BResult* PTimes::reduce (OResult* a)
// {

//  if (! (a -> isPair ())) {

//     char *or = one -> representation ();
//     char *tr = two -> representation ();
//     char *r = new char [strlen (or) + strlen (tr) + strlen ("PTIMES (, )") + 1];
//     sprintf (r, "PTIMES (%s, %s)", or, tr);
//     BResult *res = new BResBB (r, a -> Copy ());
//     delete [] or;
//     delete [] tr;
//     delete [] r;
//     delete a;
//     return res;
    
//  } else {
    
//     OResPAIR *a2 = ((OResPAIR *) a);
//     BResCOMP *res = new BResCOMP
//       (one -> reduce (a2 -> One () -> Copy ()), SAND, two -> reduce (a2 -> Two () -> Copy ()));
//     delete a;
//     return res;
//  }
 
// }


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  representation
// Parameters:	()							  
//
//  "PFO: TIMES" + one -> representation () + two -> representation ()
//         								    
//////////////////////////////////////////////////////////////////////////////

char* PTimes::representation ()
{

    char *or = one -> representation ();
    char *tr = two -> representation ();
    
    char *res = new char [strlen (or) + strlen (tr) + strlen ("( TIMES )") + 1];
    sprintf (res, "(%s TIMES %s)", or, tr);
    delete [] or;
    delete [] tr;
    return res;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transform
// Parameters:	(CState*)							  
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//
//////////////////////////////////////////////////////////////////////////////

NonType* PTimes::transform(CState* s)
{
  Predicate* first = (Predicate*) One() -> transform(s);
  Predicate* second = (Predicate*) Two() -> transform(s);
  return new PTimes(first, second,Type());
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transformDeep
// Parameters:	(CState*, CIEnvStackDataType*)							  
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//  The difference between this method and transform method is a search
//  method used in VAR classes. (Take a look at transform and transformDeep
//  methods in BVar, FVar, PVar, or OVar.
//
//////////////////////////////////////////////////////////////////////////////

NonType* PTimes::transformDeep(CState* s, CIEnvStackDataType* startingPt)
{
  Predicate* first = (Predicate*) One() -> transformDeep(s, startingPt);
  Predicate* second = (Predicate*) Two() -> transformDeep(s, startingPt);
  return new PTimes(first,second,Type());
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: compares
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for rule matching only.
// Rule matching is different from statement matching.
// Duplicated variable entry is allowed here. 
//
//////////////////////////////////////////////////////////////////////////////

int PTimes::compares(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();
  
  if ((tmp_name->tag == PNAME) && (tmp_name->pn == PRED_TIMES))
	{
	  // do it for first child.
	  if (one->compares(((PTimes*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one))
			// child (one) is a variable.
			{
			  if ((a->IFind(((PVar*)one)->Name())) == NULL)
				{
				  // child is a new variable.
				  a->IInsert(((PVar*)one)->Name(), new CIdent((PTimes*)nt,((PTimes*)nt)->One(),1));
				  // Push entry into the stack.
				}
			  else 
				{
				  // Child is a variable but not new one.
				  // In this case, I need to make sure that what's
				  // already in the stack and what I have now is the same.
				  if (!(((a->IFind(((PVar*)one)->Name()))->Child_Look_Up())->deepeq(((PTimes*)nt)->One())))
					return_val++;
				  // match failed.
				  // else do none.
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++;
	  // compare failed
	  
	  // do it for second child.
	  if (two->compares(((PTimes*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((PVar*)two)->Name())) == NULL)
				{
				  a->IInsert(((PVar*)two)->Name(), new CIdent((PTimes*)nt, ((PTimes*)nt)->Two(), 2));
				}
			  else 
				{
				  if (!(((a->IFind(((PVar*)two)->Name()))->Child_Look_Up())->deepeq(((PTimes*)nt)->Two())))
					return_val++;
				}
			}
		}
	  else return_val++;
	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0.
  // In this case, return false(0).
  // Otherwise, return true(1).
  
  delete tmp_name;
  
  if (return_val)
	return F;
  else
	return T;
  
} 

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: comparesSTM
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This match method is used for matching statement.
// No duplicated variable entry is allowed.
//
//////////////////////////////////////////////////////////////////////////////

int PTimes::comparesSTM(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();
  
  if ((tmp_name->tag == PNAME) && (tmp_name->pn == PRED_TIMES))
	{
	  // do it for first child.
	  if (one->comparesSTM(((PTimes*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one))
			// child (one) is a variable.
			{
			  if ((a->IFind(((PVar*)one)->Name())) == NULL)
				{
				  // child is a new variable.
				  a->IInsert(((PVar*)one)->Name(), new CIdent((PTimes*)nt,((PTimes*)nt)->One(),1));
				  // Push entry into the stack.
				}
			  else
				{
				  return_val++;
				  cout << "Error : Variable " << ((PVar*)one)->Name() << " is used twice. " << endl;
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++;
	  // compare failed
	  
	  // do it for second child.
	  if (two->comparesSTM(((PTimes*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((PVar*)two)->Name())) == NULL)
				{
				  a->IInsert(((PVar*)two)->Name(), new CIdent((PTimes*)nt, ((PTimes*)nt)->Two(), 2));
				}
			  else 
				{
				  return_val++;
				  cout << "Error : Variable " << ((PVar*)two)->Name() << " is used twice. " << endl;
				}
			}
		}
	  else return_val++;
	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0.
  // In this case, return false(0).
  // Otherwise, return true(1).
  
  delete tmp_name;
  
  if (return_val)
	return F;
  else
	return T;
  
} 

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  setDEPnInvDEP
// Parameters:  (char*, CState*)
//
// set dependency list and inverse dependency list
//
//////////////////////////////////////////////////////////////////////////////

void PTimes::setDEPnInvDEP(char* vn, CState* s)
{
  one->setDEPnInvDEP(vn,s);
  two->setDEPnInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  undoSetInvDEP
// Parameters:  (char*, CState*)
//
// undo changes made on inverse dependency list.
// undo for dependency list is not necessary, since 
// any changes on dependency list is undone by poping up
// top frame from the environment stack.
//
//////////////////////////////////////////////////////////////////////////////

void PTimes::undoSetInvDEP(char* vn, CState* s)
{
  one->undoSetInvDEP(vn,s);
  two->undoSetInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  replaceChild
// Parameters:	(NonType*, int)
//
// Replace a child of given tree.  
// which_child specifies which child to replace.
// For example, when which_child is one, we replace the first child of 
// given tree.
// 
//////////////////////////////////////////////////////////////////////////////

void PTimes::replaceChild(NonType* child, int which_child)
{
  if ( which_child == 1 )
	SetOne(child);
  else if ( which_child == 2 )
	SetTwo(child);
}

#ifdef _OSTORE_
//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: invoke
//  Parameters:(koObject*)
//
//  Implements the pairwise applied predicate former for koObjects.
//
//////////////////////////////////////////////////////////////////////////////

koBool* PTimes::invoke(koObject* x)
{
  return new BInvoke(this,x);
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: exec
//  Parameters:(OSObject*)
//
//  Implements the pairwise applied predicate former for OSObjects.
//
//////////////////////////////////////////////////////////////////////////////

OSBool* PTimes::exec(OSObject* x)
{
  if (((OSBPrimitive*)One()->exec(((OSPair*)x)->One()))->Val() &&
      ((OSBPrimitive*)Two()->exec(((OSPair*)x)->Two()))->Val())
    return new OSBPrimitive(1);
  else
    return new OSBPrimitive(0);
}

#endif // _OSTORE_
  
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  prolog
// Parameters:	()
//
// return string representation in prolog term
//
//////////////////////////////////////////////////////////////////////////////

char* PTimes::prolog ()
{
  char *or = one -> prolog ();
  char *tr = two -> prolog ();
  char *res = new char [strlen (or) + strlen (tr) + strlen ("ptimes(, )") + 1];
  sprintf (res, "ptimes(%s, %s)", or, tr);
  delete [] or;
  delete [] tr;
  return res;
}

#ifdef _PROLOG_
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  sp_prolog
// Parameters:	()
//
// Since C/C++ functions only have indirect access to Prolog terms 
// via SP_term_refs, we need to convert kola tree to SP_term_refs.
// Converted kola tree will get passed to prolog runtime system.
//
// See the Sicstus prolog version 3#6 manual for more detailed info on
// the usage of SP_term_refs and other prolog supplied methods.
//
//////////////////////////////////////////////////////////////////////////////

SP_term_ref PTimes::sp_prolog ()
{
  SP_term_ref or = one -> sp_prolog ();
  SP_term_ref tr = two -> sp_prolog ();
  SP_term_ref res = SP_new_term_ref();
  SP_cons_functor(res, SP_atom_from_string("ptimes"),2,or,tr);
  return res;
}
#endif // _PROLOG_

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  inv
// Parameters:	(koObject *o)
//
// Symbolic invocation of this on o.  Eventually,
// this method will replace "invoke" presently
// only defined with _OSTORE_ flag set
//
/////////////////////////////////////////////////////////////////////////////

koBool *PTimes::inv (koObject *o) 
{
  NameUnion *nu = o -> Myname ();

  if (nu ->on == OBJ_PAIR) {
	delete nu;
	return new BAnd (one -> inv (o -> asPair () -> One ()), two -> inv (o -> asPair () -> Two ()));
  } else {
	delete nu;
	cerr << "Invoking PTimes on Non-Pair Object" << endl;
    return NULL;
  }
}
