/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Mitch Cherniack (mfc@cs.brown.edu)					 	 
// FILE: TYTError.C
// PROJ: KOLA Parser, AQUA Translator
// DATE: 5/95
//
//////////////////////////////////////////////////////////////////////////////

#include "TYTError.H"


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  TError
// Parameters:	(char* n)							  
//
//  Constructor.  n is the error message
//         								    
//////////////////////////////////////////////////////////////////////////////

TError::TError (char* n)
{

    msg = new char [strlen (n) + 1];
    sprintf (msg, "%s", n);
    myname = TYPE_ERROR;
    
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  TError
// Parameters:	(char* n, TObj *t1)							  
//
//  Constructor.  n is the error message
//         								    
//////////////////////////////////////////////////////////////////////////////

TError::TError (char* n, TObj *t1)
{

    char *tr = t1 -> representation ();
    msg = new char [strlen (n) + strlen (tr) + strlen (":\n  Type Arg #1: ") + 1];
    sprintf (msg, "%s:\n  Type Arg #1: %s", n, tr);
    delete [] tr;
    myname = TYPE_ERROR;
    
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  TError
// Parameters:	(char* n, TObj *t1, TObj *t2)							  
//
//  Constructor.  n is the error message
//         								    
//////////////////////////////////////////////////////////////////////////////

TError::TError (char* n, TObj *t1, TObj *t2)
{

    char *tr = t1 -> representation ();
    char *tr2 = t2 -> representation ();
    msg = new char [strlen (n) + strlen (tr) + strlen (tr2) + strlen (":\n  Type Arg #1: \n  Type Arg #2: ") + 1];
    sprintf (msg, "%s:\n  Type Arg #1: %s\n  Type Arg #2: %s", n, tr, tr2);
    delete [] tr;
    delete [] tr2;
    myname = TYPE_ERROR;
    
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  TError
// Parameters:	(TError& other)							  
//
//  Copy Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

TError::TError (TError& other)
{

    msg = new char [strlen (other . Msg ()) + 1];
    sprintf (msg, "%s", other . Msg ());
    myname = TYPE_ERROR;
    
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~TError
// Parameters:	()
//
//  Destructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

TError::~TError ()
{
    
    delete [] msg;
    
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  representation
// Parameters:	()							  
//
//  return "ERROR" + msg
//         								    
//////////////////////////////////////////////////////////////////////////////

char* TError::representation ()
{
    
    char *res = new char [strlen (msg) + 1];
    sprintf (res, "%s", msg);
    return res;
    
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  prolog
// Parameters:	()							  
//
//         								    
//////////////////////////////////////////////////////////////////////////////

char* TError::prolog ()
{
    
  return NULL;
    
}


#ifdef _PROLOG_
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  sp_prolog
// Parameters:	()
//
// Since C/C++ functions only have indirect access to Prolog terms 
// via SP_term_refs, we need to convert kola tree to SP_term_refs.
// Converted kola tree will get passed to prolog runtime system.
//
// See the Sicstus prolog version 3#6 manual for more detailed info on
// the usage of SP_term_refs and other prolog supplied methods.
//
//////////////////////////////////////////////////////////////////////////////

SP_term_ref TError::sp_prolog ()
{
  return NULL;
}
#endif // _PROLOG_

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Copy
// Parameters:	()							  
//
//  return copy of this as TObj*
//         								    
//////////////////////////////////////////////////////////////////////////////

TObj* TError::Copy ()
{

    return new TError (*this);

}










